#define JPEG_INTERNALS
#include <math.h>
#include "jinclude.h"
#include "jpeglib.h"
#include "jdrest.h"

/* Implementation of the method developed by Aria Nosratinia:
     Enhancement of JPEG-Compressed Images by Re-application of JPEG
   (c) 2002 Kluwer Academic Publishers.

   Notes:
   - The implementation from Nikhil Hegde has been a reference for writting
     this one.
*/

#define MAXSAMP 255.
#define MINSAMP 0.

LOCAL(void)
nosratinia_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
METHODDEF(void) nosratinia_decompress_plane
  JPP((j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_buf, JBLOCKARRAY blockarray));
LOCAL(void)
fdct_inverse_image_shift(RCOEFBUFFER coefdct, RSAMPBUFFER out,
    const int w, int const h, const int shift_x, const int shift_y);
LOCAL(void)
fdct_image_shift (RCOEFBUFFER coefdct, RSAMPBUFFER in,
    const int w, int const h, const int shift_x, const int shift_y);
LOCAL(void)
average_bitmap(struct jpeg_restoration *rest);
LOCAL(void)
add2sums(struct jpeg_restoration *rest, RSAMPBUFFER from);
LOCAL(void)
loopshifts(struct jpeg_restoration *rest);
LOCAL(void)
calculate_shifts(int *shifts, int *min_shift, int *max_shift);
LOCAL(void)
project(struct jpeg_restoration *rest,
  RSAMPBUFFER out);
LOCAL(void)
quantize(struct jpeg_restoration *rest);

/* DCT information */
static const double aanscalefactor[DCTSIZE] = 
  {
    1.0, 1.387039845, 1.306562965, 1.175875602,
    1.0, 0.785694958, 0.541196100, 0.275899379
  };

enum {
  MAXSHIFTS = 64
};

struct my_rest_struct
{
  struct jpeg_restoration pub;
  RSAMPBUFFER bitmap;
  RSAMPBUFFER reference;
  RBLOCK qtable, qtable_factors, qtable_factored;
  RSAMPBUFFER sums;
  int nsums;

  RCOEFBUFFER DCTmin, DCTmax, DCTimage;
  int width_in_blocks, height_in_blocks;
  int w, h; /* _blocks * 8 */

  /* Parameters of the algorithm */
  int shifts;
  int min_shift, max_shift;
};

typedef struct my_rest_struct * my_rest_ptr;


GLOBAL(struct jpeg_restoration *)
jpeg_rest_nosratinia(j_decompress_ptr cinfo, J_REST_TARGET target)
{
  my_rest_ptr rest;

  if (target != JREST_TARGET_IDCT_Y &&
    target != JREST_TARGET_IDCT_Cb &&
    target != JREST_TARGET_IDCT_Cr)
  {
    ERREXIT(cinfo, JERR_REST_NOT_APPLICABLE);
  }

  /* The first restoration */
  rest = (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
                                SIZEOF(struct my_rest_struct));

  rest->pub.decompress_plane = nosratinia_decompress_plane;
  rest->pub.target = target;
  add_restorator(cinfo, (struct jpeg_restoration *) rest);
  return (struct jpeg_restoration *) rest;
}

METHODDEF(void)
nosratinia_decompress_plane(j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_ptr, JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,l;
  int buffer_row, block_rows;
  jpeg_component_info * compptr;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION output_col;
  RSAMP value;

  compptr = &cinfo->comp_info[myrest->pub.component];

  if (cinfo->output_iMCU_row == 0)
  {
    nosratinia_restoration(cinfo, rest, blockarray);
  }

  /* Give the rows */

  buffer_row = cinfo->output_iMCU_row * compptr->v_samp_factor;
  
  /* How many rows we should give? */
  if (cinfo->output_iMCU_row < last_iMCU_row)
    block_rows = compptr->v_samp_factor;
  else {
    /* NB: can't use last_row_height here; it is input-side-dependent! */
    block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (block_rows == 0) block_rows = compptr->v_samp_factor;
  }

  for (j=0; j<block_rows; j++)
  {
    output_col = 0;
    for (i=0; i<compptr->width_in_blocks; i++)
    {
      for(k=0; k<8;k++)
        for(l=0; l<8;l++)
        {
          value = myrest->bitmap[((buffer_row + j)*8+k)*myrest->w + (i*8+l)]
	    + 0.5;
          if (value > MAXSAMP)
            (output_ptr[k] + output_col)[l] = MAXSAMP;
          else if (value < MINSAMP)
            (output_ptr[k] + output_col)[l] = MINSAMP;
          else
            (output_ptr[k] + output_col)[l] = value;
            
        }

      output_col += compptr->DCT_scaled_size;
    }
    output_ptr += compptr->DCT_scaled_size;
  }

}

LOCAL(void)
get_env_configuration(j_decompress_ptr cinfo, struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  char *env_var;

  myrest->shifts = MAXSHIFTS;

#ifndef NO_GETENV
  env_var = getenv("JPEG_NOSRATINIA_SHIFTS");
  if (env_var != NULL)
    sscanf(env_var, "%i", &myrest->shifts);
#endif

  calculate_shifts(&myrest->shifts, &myrest->min_shift, &myrest->max_shift);
}

/* !!! Some global functions for setting the variables are needed.
jpeg_rest_nosratinia_set_shifts(rest, int)
... */

LOCAL(void)
nosratinia_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  jpeg_component_info * compptr;

  compptr = &cinfo->comp_info[myrest->pub.component];

  /* Prepare dimensions */
  myrest->width_in_blocks = compptr->width_in_blocks;
  myrest->height_in_blocks = compptr->height_in_blocks;
  myrest->w = compptr->width_in_blocks * 8;
  myrest->h = compptr->height_in_blocks * 8;
  myrest->nsums = 0;

  /* Load the coefficients, and prepare the DCT ranges */
  set_DCT_constraint(cinfo, rest, blockarray);

  get_env_configuration(cinfo, rest);

  /* Allocate the bitmap */
  myrest->bitmap =
    (RCOEFBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate the reference */
  myrest->reference =
    (RCOEFBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate the sums */
  myrest->sums =
    (RCOEFBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Zero sums */
  memset(myrest->sums, 0, sizeof(RCOEF)*myrest->h*myrest->w);

  loopshifts(rest);

  average_bitmap(rest);
}

LOCAL(void)
average_bitmap(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int index, row, col;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      index = row * myrest->w + col; 
      myrest->bitmap[index] = myrest->sums[index] / (float) myrest->nsums ;
    }
}

LOCAL(void)
loopshifts(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int shift;

  int shift_y, shift_x;

  shift_y = myrest->min_shift;
  shift_x = myrest->min_shift;

  shift = 0;

  /* Central image. Needed for the shifts */
  fdct_inverse_image(myrest->DCTimage, myrest->bitmap, myrest->w,
      myrest->h);

  memcpy(myrest->reference, myrest->bitmap, myrest->w * myrest->h
      * sizeof(RSAMP));

  /* add into the sums */
  add2sums(rest, myrest->bitmap);
  shift++;
  
  for (; shift <= myrest->shifts; shift++)
  {
    /* Forget the central image */
    if (shift_x == 0 && shift_y == 0)
    {
      shift_x++;
      continue;
    }

    /* Reget the 'reference' over the 'bitmap'.*/
    memcpy(myrest->bitmap, myrest->reference, myrest->w * myrest->h
	* sizeof(RSAMP));

    /* Decompress the image */
    fdct_image_shift(myrest->DCTimage, myrest->bitmap, myrest->w,
	myrest->h, shift_x, shift_y);

    /* Quantize the new coef with the component qtable */
    quantize(rest);

    /* Decompress the image */
    fdct_inverse_image_shift(myrest->DCTimage, myrest->bitmap, myrest->w,
	myrest->h, shift_x, shift_y);

    /* add into the sums */
    add2sums(rest, myrest->bitmap);

    /* Calculate shift offsets */
    shift_x++;
    if (shift_x > myrest->max_shift)
    {
      shift_y++;
      shift_x = myrest->min_shift;
    }
  }

  /* Empirically, it's barely needed */
  project(rest, myrest->bitmap);

}

/* IDCT on the whole image, with shifts */

LOCAL(void)
fdct_inverse_image_shift(RCOEFBUFFER coefdct, RSAMPBUFFER out,
    const int w, int const h, const int shift_x, const int shift_y)
{
  int i,j,x,y;
  RSAMP * ptr;
  RSAMP out_gray[DCTSIZE2]; 
  int index_y, index_x;

  for (i=0; i<h; i+=DCTSIZE)
    for (j=0; j<w;j+=DCTSIZE) 
      {
	fidct_buffer(coefdct, out_gray);
	coefdct+=DCTSIZE2;
	ptr=out_gray;
	for (y=0;y<DCTSIZE;y++)
	  for (x=0;x<DCTSIZE;x++)
	  {
	    index_y = i + y + shift_y;
	    if (index_y < 0)
	      index_y = 0;
	    else if (index_y >= h)
	      index_y = h - 1;

	    index_x = j + x + shift_x;
	    if (index_x < 0)
	      index_x = 0;
	    else if (index_x >= w)
	      index_x = w - 1;
	    out[index_y*w + index_x]=*(ptr++);
	  }
      }
}

/* DCT on the whole image, with shifts */

LOCAL(void)
fdct_image_shift (RCOEFBUFFER coefdct, RSAMPBUFFER in,
    const int w, int const h, const int shift_x, const int shift_y)
{
  int i,j,x,y;
  float in_gray[DCTSIZE2];
  RSAMP * ptr;
  int index_y, index_x;

  for (i=0; i<h; i+=DCTSIZE)
    for (j=0; j<w;j+=DCTSIZE) 
    {
      ptr=in_gray;
      for (y=0;y<DCTSIZE;y++)
	for (x=0;x<DCTSIZE;x++)
	{
	  index_y = i + y + shift_y;
	  if (index_y < 0)
	    index_y = 0;
	  else if (index_y >= h)
	    index_y = h - 1;

	  index_x = j + x + shift_x;
	  if (index_x < 0)
	    index_x = 0;
	  else if (index_x >= w)
	    index_x = w - 1;

	  *(ptr++)=in[index_y*w + index_x];
	}
      fdct_buffer(in_gray, coefdct);
      coefdct+=DCTSIZE2;
    }
}


LOCAL(void)
quantize(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i;
  int num_coef = 0;
  RCOEFBUFFER coef = myrest->DCTimage;

  for (i=0; i<myrest->height_in_blocks * myrest->width_in_blocks*DCTSIZE2; ++i)
  {
    num_coef = i % DCTSIZE2;
    *coef = *coef / myrest->qtable_factored[num_coef];
    *coef += 0.5;
    *coef = floor(*coef) * myrest->qtable_factored[num_coef];
    coef++;
  }
}

LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,row,col, adr;
  RCOEF *table;

  myrest->DCTmin =
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTmax = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTimage = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);

  if ( (myrest->DCTmin == NULL) || (myrest->DCTmax == NULL) ||
      (myrest->DCTimage == NULL) )
    ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 0);
  
  /* Get quantization table */
  for (i=0; i < DCTSIZE2; i++)
    myrest->qtable[i] = cinfo->comp_info[myrest->pub.component].quant_table->quantval[i];

  table = myrest->qtable_factored;
 
  /* For float AA&N IDCT method, multipliers are equal to quantization
   * coefficients scaled by scalefactor[row]*scalefactor[col], where
   *   scalefactor[0] = 1
   *   scalefactor[k] = cos(k*PI/16) * sqrt(2)    for k=1..7
   */
  i = 0;
  for (row = 0; row < DCTSIZE; row++) 
    for (col = 0; col < DCTSIZE; col++) 
      {
	myrest->qtable_factors[i] = aanscalefactor[row] * aanscalefactor[col];
        table[i]=myrest->qtable[i] * myrest->qtable_factors[i];
        i++;
      }
  
  for (j=0;j<myrest->height_in_blocks;j++)
    for (i=0;i<myrest->width_in_blocks;i++)
      for (k=0;k<DCTSIZE2;k++)
      {
        adr = (j*myrest->width_in_blocks + i)*DCTSIZE2 + k;
        myrest->DCTmin[adr]=((RCOEF) blockarray[j][i][k]-0.5)*table[k];
        myrest->DCTmax[adr]=((RCOEF) blockarray[j][i][k]+0.5)*table[k];
        myrest->DCTimage[adr]=(myrest->DCTmin[adr]+myrest->DCTmax[adr])/2.;
      }
}

LOCAL(void)
calculate_shifts(int *shifts, int *min_shift, int *max_shift)
{
  if (*shifts < 1)
    *shifts = 1;
  else if (*shifts > MAXSHIFTS)
    *shifts = MAXSHIFTS;

  if (*shifts == 1)
  {
    *min_shift = 0;
    *max_shift = 0;
  }
  else if (*shifts <= 3*3)
  {
    *min_shift = -1;
    *max_shift = 1;
  }
  else if (*shifts <= 5*5)
  {
    *min_shift = -2;
    *max_shift = 2;
  }
  else if (*shifts <= 7*7)
  {
    *min_shift = -3;
    *max_shift = 3;
  }
  else if (*shifts <= 8*8)
  {
    *min_shift = -3;
    *max_shift = 4;
  }
}

LOCAL(void)
add2sums(struct jpeg_restoration *rest, RSAMPBUFFER from)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int row,col, index;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      index = row * myrest->w + col; 
      myrest->sums[index] += from[index];
    }
  /*
  fprintf(stderr, "from[8000]= %f\t", from[8000]);
  fprintf(stderr, "from[8001]= %f\n", from[8001]);
  */
  
  ++myrest->nsums;
}


/* Projection on the constraint - by Froment*/
LOCAL(void)
project(struct jpeg_restoration *rest,
  RSAMPBUFFER out)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;

  int j;
  fdct_image(myrest->DCTimage, out, myrest->w, myrest->h);
    
  for (j=0;j<myrest->h*myrest->w;j++)
    if (myrest->DCTimage[j] > myrest->DCTmax[j])
      myrest->DCTimage[j] = myrest->DCTmax[j];
    else if (myrest->DCTimage[j] < myrest->DCTmin[j])
      myrest->DCTimage[j] = myrest->DCTmin[j];

  fdct_inverse_image(myrest->DCTimage, out, myrest->w, myrest->h);

}
